from typing import Any, Literal

from pydantic import Field

from abuse_whois.schemas.whois import WhoisRecord
from abuse_whois.utils import is_email, is_url

from .api_model import APIModel


class Contact(APIModel):
    provider: str = Field(..., description="Provider name")
    address: str = Field(..., description="Contact address")
    type: Literal["email", "form"] = Field(
        "email", description="Type of contact method"
    )

    def __init__(self, **data: Any):
        super().__init__(**data)

        if is_email(self.address):
            self.type = "email"

        if is_url(self.address):
            self.type = "form"


class Contacts(APIModel):
    address: str
    hostname: str = Field(..., description="Host name")

    ip_address: str | None = Field(None, description="IP address")
    registered_domain: str | None = Field(
        None, description="Registered domain (a.k.a. free level domain)"
    )

    shared_hosting_provider: Contact | None = Field(
        None, description="Shared hosting provider"
    )
    registrar: Contact | None = Field(None, description="Registrar")
    hosting_provider: Contact | None = Field(None, description="Hosting provider")

    whois_record: WhoisRecord | None = Field(
        None, description="Whois record of hostname"
    )



import pathlib
from functools import lru_cache

from ... import settings
from ...utils import glob_rules, load_yaml
from .rule import WhoisRule

DEFAULT_RULE_DIRECTORY: pathlib.Path = pathlib.Path(__file__).parent / "./rules"


@lru_cache(maxsize=1)
def load_rules() -> list[WhoisRule]:
    paths = glob_rules(
        DEFAULT_RULE_DIRECTORY,
        additional_directories=settings.ADDITIONAL_WHOIS_RULE_DIRECTORIES,
    )
    return [WhoisRule.model_validate(load_yaml(path)) for path in paths]



import asyncio
import json

import typer

from . import schemas
from .errors import InvalidAddressError
from .matchers.shared_hosting import get_shared_hosting_provider
from .matchers.whois import get_optional_whois_contact
from .utils import (
    get_hostname,
    get_registered_domain,
    is_domain,
    is_ip_address,
    is_supported_address,
    resolve,
)
from .whois import get_whois_record


async def get_abuse_contacts(address: str) -> schemas.Contacts:
    if not is_supported_address(address):
        raise InvalidAddressError(f"{address} is not supported type address")

    hostname = get_hostname(address)  # Domain or IP address

    domain: str | None = None
    ip_address: str | None = None
    registered_domain: str | None = None

    if is_domain(hostname):
        domain = hostname
        registered_domain = get_registered_domain(hostname)

        # get IP address by domain
        try:
            ip_address = await resolve(hostname)
        except OSError:
            pass

    if is_ip_address(hostname):
        ip_address = hostname

    whois_record = await get_whois_record(hostname)
    shared_hosting_provider = get_shared_hosting_provider(hostname)

    registrar, hosting_provider = await asyncio.gather(
        get_optional_whois_contact(domain), get_optional_whois_contact(ip_address)
    )

    return schemas.Contacts(
        address=address,
        hostname=hostname,
        ip_address=ip_address,
        registered_domain=registered_domain,
        shared_hosting_provider=shared_hosting_provider,
        registrar=registrar,
        hosting_provider=hosting_provider,
        whois_record=whois_record,
    )


app = typer.Typer()


@app.command()
def whois(
    address: str = typer.Argument(..., help="URL, domain, IP address or email address")
):
    try:
        contacts = asyncio.run(get_abuse_contacts(address))
        print(contacts.model_dump_json(by_alias=True))  # noqa: T201
    except (InvalidAddressError, asyncio.TimeoutError) as e:
        print(json.dumps({"error": str(e)}))  # noqa: T201


if __name__ == "__main__":
    app()



from humps import camelize
from pydantic import BaseModel, ConfigDict


class APIModel(BaseModel):
    model_config = ConfigDict(
        from_attributes=True, alias_generator=camelize, populate_by_name=True
    )



import asyncio

from fastapi import APIRouter, HTTPException, status

from abuse_whois import get_abuse_contacts, schemas
from abuse_whois.errors import InvalidAddressError

router = APIRouter(prefix="/whois")


@router.post("/", response_model=schemas.Contacts)
async def whois(query: schemas.Query) -> schemas.Contacts:
    try:
        return await get_abuse_contacts(query.address)
    except InvalidAddressError as e:
        raise HTTPException(status.HTTP_400_BAD_REQUEST, detail=str(e)) from e
    except asyncio.TimeoutError as e:
        raise HTTPException(status.HTTP_408_REQUEST_TIMEOUT, detail=str(e)) from e



import sys

from starlette.config import Config
from starlette.datastructures import CommaSeparatedStrings

config = Config(".env")

# FastAPI settings

PROJECT_NAME: str = config("PROJECT_NAME", default="abuse-whois")

DEBUG: bool = config("DEBUG", cast=bool, default=False)
TESTING: bool = config("TESTING", cast=bool, default=False)

LOG_FILE = config("LOG_FILE", default=sys.stderr)
LOG_LEVEL: str = config("LOG_LEVEL", cast=str, default="DEBUG")
LOG_BACKTRACE: bool = config("LOG_BACKTRACE", cast=bool, default=True)

# Abuse Whois setting
WHOIS_LOOKUP_MAX_RETRIES: int = config("WHOIS_LOOKUP_MAX_RETRIES", cast=int, default=3)
WHOIS_LOOKUP_TIMEOUT: int = config("WHOIS_LOOKUP_TIMEOUT", cast=int, default=10)
WHOIS_LOOKUP_CACHE_SIZE: int = config("WHOIS_LOOKUP_CACHE_SIZE", cast=int, default=1024)
WHOIS_LOOKUP_CACHE_TTL: int = config(
    "WHOIS_LOOKUP_CACHE_TTL", cast=int, default=60 * 60
)

IP_ADDRESS_LOOKUP_TIMEOUT: int = config(
    "IP_ADDRESS_LOOKUP_TIMEOUT", cast=int, default=10
)
IP_ADDRESS_LOOKUP_CACHE_SIZE: int = config(
    "IP_ADDRESS_LOOKUP_CACHE_SIZE", cast=int, default=1024
)
IP_ADDRESS_LOOKUP_CACHE_TTL: int = config(
    "IP_ADDRESS_LOOKUP_CACHE_TTL", cast=int, default=60 * 60
)

RULE_EXTENSIONS: CommaSeparatedStrings = config(
    "RULE_EXTENSIONS", cast=CommaSeparatedStrings, default="yaml,yml"
)

ADDITIONAL_WHOIS_RULE_DIRECTORIES: CommaSeparatedStrings = config(
    "ADDITIONAL_WHOIS_RULE_DIRECTORIES", cast=CommaSeparatedStrings, default=""
)

ADDITIONAL_SHARED_HOSTING_RULE_DIRECTORIES: CommaSeparatedStrings = config(
    "ADDITIONAL_SHARED_HOSTING_RULE_DIRECTORIES", cast=CommaSeparatedStrings, default=""
)



from ...schemas import BaseRule
from ...whois import get_whois_record


class WhoisRule(BaseRule):
    async def match(self, hostname: str) -> bool:
        whois_record = await get_whois_record(hostname)
        data = whois_record.model_dump(by_alias=True)
        return super().match(data)



import asyncio
import pathlib
import socket
from collections.abc import Callable
from contextlib import contextmanager
from functools import lru_cache
from typing import cast
from urllib.parse import urlparse

import tldextract
import validators
import yaml
from asyncer import asyncify
from cachetools import TTLCache, cached
from starlette.datastructures import CommaSeparatedStrings

from . import settings


def _is_x(v: str, *, validator: Callable[[str], bool]) -> bool:
    res = validator(v)

    if isinstance(res, validators.ValidationError):
        return False

    return res


def is_ip_address(v: str) -> bool:
    return _is_x(v, validator=validators.ipv4) or _is_x(v, validator=validators.ipv6)  # type: ignore


def is_domain(v: str) -> bool:
    if len(v.split(".")) == 1:
        return False

    if "@" in v:
        return False

    return _is_x(v, validator=validators.domain)  # type: ignore


def is_url(v: str) -> bool:
    if not v.startswith(("http://", "https://")):
        return False

    return _is_x(v, validator=validators.url)  # type: ignore


def is_email(v: str) -> bool:
    return _is_x(v, validator=validators.email)  # type: ignore


def is_supported_address(v: str) -> bool:
    if is_domain(v) or is_ip_address(v) or is_email(v) or is_url(v):
        return True

    return False


@lru_cache(maxsize=settings.WHOIS_LOOKUP_CACHE_SIZE)
def get_registered_domain(v: str) -> str | None:
    parsed = tldextract.extract(v)

    if parsed.registered_domain == "":
        return None

    return parsed.registered_domain


def get_hostname(value: str) -> str:
    if is_ip_address(value) or is_domain(value):
        return value

    if is_email(value):
        value = f"http://{value}"

    parsed = urlparse(value)
    return parsed.hostname or value


@contextmanager
def with_socket_timeout(timeout: float):
    old = socket.getdefaulttimeout()
    try:
        socket.setdefaulttimeout(timeout)
        yield
    except (socket.timeout, ValueError) as e:
        raise asyncio.TimeoutError(
            f"{timeout} seconds have passed but there is no response"
        ) from e
    finally:
        socket.setdefaulttimeout(old)


@cached(
    cache=TTLCache(
        maxsize=settings.IP_ADDRESS_LOOKUP_CACHE_SIZE,
        ttl=settings.IP_ADDRESS_LOOKUP_CACHE_TTL,
    )
)
def _resolve(
    hostname: str, *, timeout: float = float(settings.IP_ADDRESS_LOOKUP_TIMEOUT)
) -> str:
    with with_socket_timeout(timeout):
        ip = socket.gethostbyname(hostname)
        return ip


resolve = asyncify(_resolve)


def load_yaml(path: str | pathlib.Path) -> dict:
    with open(path) as f:
        return cast(dict, yaml.safe_load(f))


def glob_rules(
    base_directory: str | pathlib.Path,
    *,
    additional_directories: list[str] | list[pathlib.Path] | CommaSeparatedStrings,
    rule_extensions=settings.RULE_EXTENSIONS,
) -> list[pathlib.Path]:
    directories = [base_directory]
    directories.extend(additional_directories)

    directories = [pathlib.Path(d) for d in directories]

    paths: set[str] = set()
    for directory in directories:
        for extension in rule_extensions:
            paths.update([str(p) for p in directory.glob(f"*.{extension}")])  # type: ignore

    return [pathlib.Path(p) for p in paths]



import re

from abuse_whois.schemas import Contact, WhoisRecord
from abuse_whois.utils import is_email
from abuse_whois.whois import get_whois_record

from .rules import load_rules


def get_whois_abuse_contact_by_regexp(
    record: WhoisRecord,
    *,
    email_pattern: str = r"[a-z0-9\.\-+_]+@[a-z0-9\.\-+_]+\.[a-z]+"
) -> str | None:
    matches = re.findall(email_pattern, record.raw_text)
    if len(matches) == 0:
        return None

    # returns the email address which contains "abuse"
    emails = [str(match) for match in matches if is_email(str(match))]
    emails.reverse()

    for email in emails:
        if "abuse" in email:
            return email

    return None


def get_whois_abuse_contact(record: WhoisRecord) -> Contact | None:
    provider = record.registrar
    email: str | None = None

    # check email format for just in case
    if is_email(record.abuse.email or ""):
        email = record.abuse.email

    if email is None:
        # fallback to regexp based search
        email = get_whois_abuse_contact_by_regexp(record)
        if email is None:
            return None

    # use email's domain as a provider if provider is None
    if provider is None and is_email(email or ""):
        provider = email.split("@")[-1]

    if provider is None or email is None:
        return None

    return Contact(provider=provider, address=email)


async def get_whois_contact(
    hostname: str,
) -> Contact | None:
    rules = load_rules()
    for rule in rules:
        if await rule.match(hostname):
            return rule.contact

    # Use whois registrar & abuse data
    try:
        whois_record = await get_whois_record(hostname)
    except Exception:
        return None

    return get_whois_abuse_contact(whois_record)


async def get_optional_whois_contact(hostname: str | None) -> Contact | None:
    if hostname is None:
        return None

    return await get_whois_contact(hostname)



from abuse_whois.schemas import Contact

from .rules import load_rules


def get_shared_hosting_provider(hostname: str) -> Contact | None:
    rules = load_rules()
    for rule in rules:
        if rule.match(hostname):
            return rule.contact

    return None



from fastapi import FastAPI
from loguru import logger

from abuse_whois import settings

from .endpoints import index, whois


def create_app():
    logger.add(
        settings.LOG_FILE, level=settings.LOG_LEVEL, backtrace=settings.LOG_BACKTRACE
    )

    app = FastAPI(
        debug=settings.DEBUG,
        title=settings.PROJECT_NAME,
    )

    # add routes
    app.include_router(index.router)
    app.include_router(whois.router, prefix="/api")

    return app


app = create_app()



import pathlib
from functools import lru_cache

from abuse_whois import settings
from abuse_whois.utils import glob_rules, load_yaml

from .rule import SharedHostingRule

DEFAULT_RULE_DIRECTORY: pathlib.Path = pathlib.Path(__file__).parent / "./rules"


@lru_cache(maxsize=1)
def load_rules() -> list[SharedHostingRule]:
    paths = glob_rules(
        DEFAULT_RULE_DIRECTORY,
        additional_directories=settings.ADDITIONAL_SHARED_HOSTING_RULE_DIRECTORIES,
    )
    return [SharedHostingRule.model_validate(load_yaml(path)) for path in paths]



import asyncio
from dataclasses import asdict

import stamina
from asyncache import cached
from asyncwhois.query import DomainQuery, NumberQuery
from cachetools import TTLCache
from whois_parser import WhoisParser, WhoisRecord

from . import schemas, settings
from .errors import RateLimitError
from .utils import get_registered_domain, is_domain, is_ip_address

whois_parser = WhoisParser()


def parse(
    raw_text: str, hostname: str, *, parser: WhoisParser = whois_parser
) -> WhoisRecord:
    return parser.parse(raw_text, hostname=hostname)


async def query(address: str, *, timeout: int = settings.WHOIS_LOOKUP_TIMEOUT) -> str:
    klass = DomainQuery if is_domain(address) else NumberQuery
    query = await klass.new_aio(address, timeout=timeout)
    return query.query_output


@stamina.retry(
    on=(asyncio.TimeoutError, RateLimitError),
    attempts=settings.WHOIS_LOOKUP_MAX_RETRIES,
    timeout=None,
)
@cached(
    cache=TTLCache(
        maxsize=settings.WHOIS_LOOKUP_CACHE_SIZE, ttl=settings.WHOIS_LOOKUP_CACHE_TTL
    )
)
async def get_whois_record(
    hostname: str,
    *,
    timeout: int = settings.WHOIS_LOOKUP_TIMEOUT,
    parser: WhoisParser = whois_parser
) -> schemas.WhoisRecord:
    if not is_ip_address(hostname):
        hostname = get_registered_domain(hostname) or hostname

    query_result = await query(hostname, timeout=timeout)
    query_result = "\n".join(query_result.splitlines())

    parsed = parse(query_result, hostname, parser=parser)
    if parsed.is_rate_limited:
        raise RateLimitError()

    return schemas.WhoisRecord.model_validate(asdict(parsed))



from datetime import datetime

from pydantic import Field

from .api_model import APIModel


class Contact(APIModel):
    organization: str | None = None
    email: str | None = None
    name: str | None = None
    telephone: str | None = None


class Abuse(APIModel):
    email: str | None = None
    telephone: str | None = None


class WhoisRecord(APIModel):
    raw_text: str

    tech: Contact
    admin: Contact
    registrant: Contact
    abuse: Abuse

    statuses: list[str] = Field(default_factory=list)
    name_servers: list[str] = Field(default_factory=list)

    domain: str | None = None
    registrar: str | None = None

    expires_at: datetime | str | None = None
    registered_at: datetime | str | None = None
    updated_at: datetime | str | None = None
